package cz.upol.foolang.jbc;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtField;
import javassist.NotFoundException;
import javassist.bytecode.BadBytecode;

/**
 * Trida poskytujici rozhrani pro pohodlnejsi generovani trid
 * pomoci knihovny JavaAssist
 */
public class ClassBuilder {
	
	private final ClassPool classPool = ClassPool.getDefault();

	/** vysledna generovana trida */
	private final CtClass generatedClass;
	
	/** objekty, ktere byly pouzity k sestaveni metod */
	private final List<MethodBuilder> methodBuilders = new ArrayList<>();
	
	public ClassBuilder(String className) {
		generatedClass = classPool.makeClass(className);
	}
	
	/** 
	 * Vraci objekt (builder), ktery lze pouzit k vytvoreni metody
	 * 
	 * @param name -- nazev metody
	 * @param modifiers -- modifikatory {@link javaasist.Modifier}
	 * @param returnType -- typ navratove hodnoty, pro popis se pouzivaji tridy CtClass z knihovny JavaAssist
	 * @param paramTypes... -- parametry metody  
	 */
	public MethodBuilder createMethod(String name, int modifiers, CtClass returnType, CtClass... paramTypes) {
		MethodBuilder methodBuilder = new MethodBuilder(generatedClass, name, modifiers, returnType, paramTypes);
		methodBuilders.add(methodBuilder);
		return methodBuilder;
	}
	
	/**
	 * Prida do dane tridy atribut
	 * 
	 * @param fieldClass -- typ attribut, pro popis se pouzivaji tridy CtClass z knihovny JavaAssist
	 * @param fieldName -- nazev atributu
	 * @param modifiers -- modifikatory {@link javaasist.Modifier}
	 */
	public void addField(CtClass fieldClass, String fieldName, int modifiers) throws CannotCompileException {
		CtField field = new CtField(fieldClass, fieldName, generatedClass);
		field.setModifiers(modifiers);
		generatedClass.addField(field);
	}
	
	/**
	 * Vygeneruje a ulozi do souboru prislusnou tridu
	 */
	public void save(String directory) throws NotFoundException, IOException, CannotCompileException, BadBytecode {
			
		for (MethodBuilder methodBuilder : methodBuilders) {
			methodBuilder.save();
			generatedClass.addMethod(methodBuilder.getMethod());
		}
		generatedClass.writeFile(directory);
	}
	
	public void setSuperClass(CtClass superClass) throws CannotCompileException {
		generatedClass.setSuperclass(superClass);
	}
	
	public void addInterface(CtClass iface) {
		generatedClass.addInterface(iface);
	}
	
	public ClassPool getClassPool() {
		return classPool;
	}
}
