package cz.upol.foolang.jbc;

import javassist.CtClass;
import javassist.CtMethod;
import javassist.bytecode.BadBytecode;
import javassist.bytecode.Bytecode;
import javassist.bytecode.ConstPool;
import javassist.bytecode.MethodInfo;

public class MethodBuilder {

	private final CtClass generatedClass;
	private final CtMethod method;
	private final MethodInfo methodInfo;
	private Bytecode bytecode;

	MethodBuilder(CtClass generatedClass, String methodName, int modifiers, CtClass returnType, CtClass[] paramTypes) {
		this.generatedClass = generatedClass;
		
		this.method = new CtMethod(returnType, methodName, paramTypes, generatedClass);
		this.method.setModifiers(modifiers);
		this.methodInfo = this.method.getMethodInfo();
	}
	
	public Bytecode createByteCode(int stackSize, int locals) {
		if (bytecode != null) throw new RuntimeException("Bytecode already exsits");
		bytecode = new Bytecode(methodInfo.getConstPool(), stackSize, locals);
		return bytecode;
	}
	
	public void save() throws BadBytecode  {
		methodInfo.setCodeAttribute(bytecode.toCodeAttribute());
		methodInfo.rebuildStackMap(generatedClass.getClassPool());
	}

	public CtMethod getMethod() {
		return method;
	}
	
	public ConstPool getConstantPool() {
		return methodInfo.getConstPool();
	}

	public CtClass getGeneratedClass() {
		return generatedClass;
	}

}
