package cz.upol.foolang.tutorials;

import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;

import cz.upol.foolang.jbc.ClassBuilder;
import cz.upol.foolang.jbc.MethodBuilder;
import javassist.CtClass;
import javassist.Modifier;
import javassist.bytecode.Bytecode;
import javassist.bytecode.Opcode;

public class Sample01 {
	
	private static final String WORKING_DIRECTORY = System.getProperty("java.io.tmpdir");
	private static final String CLASS_NAME = "FooClass";
	
	/**
	 * Sestavi a otestuje jednoduche metody
	 */
	public static void main(String[] args) throws Exception {
		System.out.println("Working directory: " + WORKING_DIRECTORY);
			
		ClassBuilder barBuilder = new ClassBuilder(CLASS_NAME);
		createIncMethod(barBuilder);
		createIntSqrMethod(barBuilder);
		createDoubleSqrMethod(barBuilder);

		barBuilder.save(WORKING_DIRECTORY);
		
		testMethods();		
	}
	
	/**
	 * Vytvori metodu odpovidajici deklaraci:
	 * public static int inc(int a) {
	 *	 return a + 1;
	 * }
	 */
	private static void createIncMethod(ClassBuilder classBuilder) {
		MethodBuilder incMethod = classBuilder.createMethod("inc", Modifier.PUBLIC | Modifier.STATIC, CtClass.intType, CtClass.intType);
		
		Bytecode bc = incMethod.createByteCode(2, 1);
		bc.addIconst(1); // vlozi konstantu na zasobnik (pouzije nejvhodnejsi operaci)
		bc.addIload(0);  // vlozi prvni operand (pouzije nejvhodnejsi operaci)
		bc.addOpcode(Opcode.IADD);
		bc.addOpcode(Opcode.IRETURN);
	}
	
	/**
	 * Vytvori metodu odpovidajici deklaraci:
	 * public static int sqr(int a) {
	 *	 return a * a;
	 * }
	 */
	private static void createIntSqrMethod(ClassBuilder classBuilder) {
		MethodBuilder sqrMethod = classBuilder.createMethod("sqr", Modifier.PUBLIC | Modifier.STATIC, CtClass.intType, CtClass.intType);
		
		Bytecode bc = sqrMethod.createByteCode(2, 1);
		bc.addIload(0);
		bc.addOpcode(Opcode.DUP);  // zdvoji hodnotu na vrcholu zasobniku
		bc.addOpcode(Opcode.IMUL);
		bc.addOpcode(Opcode.IRETURN);
	}
	
	/**
	 * Vytvori metodu odpovidajici deklaraci:
	 * public static double sqr(double a) {
	 *	 return a * a;
	 * }
	 */
	private static void createDoubleSqrMethod(ClassBuilder classBuilder) {
		MethodBuilder sqrMethod = classBuilder.createMethod("sqr", Modifier.PUBLIC | Modifier.STATIC, CtClass.doubleType, CtClass.doubleType);
		
		Bytecode bc = sqrMethod.createByteCode(4, 2); // jedna hodnota typu double zabira dve slova 
		bc.addDload(0);
		bc.addOpcode(Opcode.DUP2);  // zdvoji hodnotu na vrcholu zasobniku (zabira dve slova)
		bc.addOpcode(Opcode.DMUL);
		bc.addOpcode(Opcode.DRETURN);
	}
	
	/**
	 * Otestuje metody vytvorene v metode {@link #main}
	 */
	public static void testMethods() throws Exception {
		
		// nacte tridu z fyzickeho souboru
		URL url = new URL("file:" + WORKING_DIRECTORY + "/");
		System.out.println(url);
		URLClassLoader cl = new URLClassLoader(new URL[] { url });
		Class<?> c = cl.loadClass(CLASS_NAME);
		
		// vyvola metodu inc s argumentem 5
		Method m1 = c.getMethod("inc", int.class);
		Object x1 = m1.invoke(null, 5);
		System.out.println(x1);
		
		// vyvola metodu sqr s argumentem 4 (int)
		Method m2 = c.getMethod("sqr", int.class);
		Object x2 = m2.invoke(null, 4);
		System.out.println(x2);
		
		// vyvola metodu sqr s argumentem 3.3 (double)
		Method m3 = c.getMethod("sqr", double.class);
		Object x3 = m3.invoke(null, 3.3);
		System.out.println(x3);
		
		cl.close();
	}
}
